<?php

namespace Drupal\web3_provider\Plugin\Web3Provider;

use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Render\Markup;
use Drupal\Core\Url;
use Drupal\web3_provider\Plugin\Web3Provider\Web3ProviderBase;
use Drupal\web3_provider\Utility\Transaction;
use Web3p\EthereumUtil\Util;

/**
 * Defines a default web3_provider plugin to communicate with ETH nodes.
 *
 * @Web3Provider(
 *   id = "custom",
 *   title = @Translation("Custom"),
 *   description = @Translation("HTTP Provider and Signer selected from secret keys.")
 * )
 */
class Custom extends Web3ProviderBase {

  use MessengerTrait;

  /**
   * {@inheritDoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $config = $this->getConfiguration();
    
    $form['chainId']['#default_value'] = $config['chainId'];

    $url_options = ['attributes' => ['target' => '_blank']];

    $form['web3_http_provider'] = [
      '#type' => 'key_select',
      '#title' => $this->t('HTTP Provider'),
      '#description' => '<br>' .
        $this->t('This is usually an <em>authentication</em> key holding a URL including your API key included.') . '<br>' .
        $this->t('Example of recommended provider: @list', [
          '@list' => Markup::create(implode(', ', [
            Link::fromTextAndUrl('Alchemy', Url::fromUri('https://www.alchemy.com/', $url_options))->toString(),
            Link::fromTextAndUrl('Infura', Url::fromUri('https://www.infura.io/', $url_options))->toString(),
            Link::fromTextAndUrl('Amazon Managed Blockchain', Url::fromUri('https://aws.amazon.com/managed-blockchain/', $url_options))->toString(),
            Link::fromTextAndUrl('cloudfare', Url::fromUri('https://www.cloudflare.com/web3/', $url_options))->toString(),
          ]))
        ]),
      '#default_value' => $config['web3_http_provider'] ?? NULL,
      '#required' => TRUE,
      // @todo Custom key type: web3_http_provider
      '#key_filters' => ['type' => 'authentication'],
    ];

    $form['web3_signer'] = [
      '#type' => 'key_select',
      '#title' => $this->t('Signer'),
      '#description' => '<br>' .
        $this->t('This is usually an <em>authentication</em> key holding your private key.'),
      '#default_value' => $config['web3_signer'] ?? NULL,
      '#required' => FALSE,
      // @todo Custom key type: web3_signer
      '#key_filters' => ['type' => 'authentication'],
    ];

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function buildTransactionForm(array $form, FormStateInterface $form_state) {
    // Stop now if transaction was submitted.
    if ($hash = $form_state->get('hash')) {
      $form['transaction'] = ['#markup' => $hash];
      return $form;
    }

    // Build default transaction form.
    $form = parent::buildTransactionForm($form, $form_state);

    if ($form_state instanceof SubformStateInterface) {
      $form_state = $form_state->getCompleteFormState();
    }

    $wrapper_id = Html::getId($form_state->getBuildInfo()['form_id']);
    $form['#id'] = $wrapper_id;

    // Default network.
    $form['chainId']['#default_value'] = $this->configuration['chainId'] ?? NULL;
    $form['chainId']['#disabled'] = TRUE;

    // Get address from private key.
    if (!($from = $form_state->getValue('from'))) {
      $key_id = $this->configuration['web3_signer'];
      $key = $this->keyRepository->getKey($key_id);
      $secrets = $key->getKeyValues();
      $private_key = reset($secrets) ?? NULL;

      $utils = new Util;
      $from = $utils->publicKeyToAddress(
        $utils->privateKeyToPublicKey($private_key)
      );

      $form['from']['#default_value'] = $from;
      $form['from']['#disabled'] = !empty($from);
    }

    // Default values.
    if ($from && !$form_state->get('set_default_values')) {
      // Get latest block info.
      $requests = [];
      $requests[] = ['eth_blockNumber' => []];
      $results = $this->requests($requests);
      $block_number = $results[0]['result'] ?? [];
      unset($requests, $results);


      // Get fees and nonce.
      $requests = [];
      $requests[] = ['eth_getBlockByNumber' => [$block_number, false]];
      $requests[] = ['eth_getTransactionCount' => [$from, 'latest']];
      $results = $this->requests($requests);

      if ($baseFeePerGas = $results[0]['result']['baseFeePerGas'] ?? NULL) {
        $form['baseFeePerGas']['#default_value'] = hexdec($baseFeePerGas);
        $form['maxFeePerGas']['#default_value'] = hexdec($baseFeePerGas) + 1;
        $form['maxPriorityFeePerGas']['#default_value'] = 1;
      }
      if ($nonce = $results[1]['result'] ?? NULL) {
        $form['nonce']['#default_value'] = hexdec($nonce);
      }

      $form_state->set('set_default_values', true);
    }

    // Actions.
    $form['actions'] = ['#type' => 'actions'];

    $form['actions']['send'] = [
      '#type' => 'submit',
      '#value' => $this->t('Send'),
      '#ajax' => [
        'callback' => '::ajaxRefresh',
        'wrapper' => $wrapper_id,
      ],
    ];

    return $form;
  }

  /**
   * Ajax callback.
   */
  public static function ajaxRefresh(array &$form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function validateTransactionForm(array &$form, FormStateInterface $form_state) {
    $key_id = $this->configuration['web3_signer'];
    if (!$key_id) {
      $form_state->setErrorByName('web3_signer', $this->t('Missing web3_signer.') . '<br>' . $this->t('Please reconfigure this Web3 provider.'));
      return;
    }

    // Retrieve and validate private key for later use in submit.
    $key = $this->keyRepository->getKey($key_id);
    $secrets = $key->getKeyValues();
    $private_key = reset($secrets) ?? NULL;
    if (!$private_key || empty($private_key)) {
      $form_state->setErrorByName('web3_signer', $this->t('Missing or empty private key'));
      return;
    }

    $transaction = new Transaction(array_filter($form_state->getValues()));


    try {
      // Test transaction by estimating cost.
      $requests = [];
      $requests[] = ['eth_estimateGas' => [$transaction->data, 'latest']];
      $results = $this->requests($requests);
      $estimated_gas = $form_state->getValue('gas') ?? $results[0]['result'] ?? NULL;
      if ($error = $results[0]['error']['message'] ?? NULL) {
        throw new \Exception($error);
      }
    } catch (\Exception $e) {
      $form_state->setErrorByName(NULL, $e->getMessage());
      return;
    }

    // Sign the transaction with private key now.
    $transaction->refresh(['gas' => $estimated_gas]);
    $form_state->set('signed_tx',  $transaction->sign($private_key));
  }

  /**
   * {@inheritDoc}
   */
  public function submitTransactionForm(array &$form, FormStateInterface $form_state) {
    $signed_tx = $form_state->get('signed_tx');

    try {
      $requests = [];
      $requests[] = ['eth_sendRawTransaction' => [$signed_tx]];
      $results = $this->requests($requests);
      $result = reset($results) ?? [];

      if ($error = $result['error']['message'] ?? NULL) {
        throw new \Exception($error);
      }

      // Save transaction hash.
      $hash = $result['result'] ?? NULL;
      $message = $this->t('Transaction sent with hash: @hash', [
        '@hash' => $hash,
      ]);
      $this->messenger()->addStatus($message);
      $this->logger->notice($message);

      $form_state->set('hash', $hash);
    } catch (\Exception $e) {
      $this->messenger()->addError($e->getMessage());
    }
  }
}
