<?php

namespace Drupal\web3_provider\Plugin\Web3Provider;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformStateInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Render\Markup;
use Drupal\web3_provider\Ajax\ScrollToCommand;
use Drupal\web3_provider\Ajax\WalletCommand;
use Drupal\web3_provider\Plugin\Web3Provider\Web3ProviderBase;
use Drupal\web3_provider\Utility\Transaction;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines the web3_provider plugin to sign a transaction with a Web3 wallet.
 *
 * @Web3Provider(
 *   id = "wallet",
 *   title = @Translation("Wallet"),
 *   description = @Translation("Interact with web3 wallets.")
 * )
 */
class Wallet extends Web3ProviderBase {

  use MessengerTrait;

  /**
   * The library discovery service.
   *
   * @var \Drupal\Core\Asset\LibraryDiscovery
   */
  protected $libraryManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->libraryManager = $container->get('library.discovery');
    return $instance;
  }

  /**
   * {@inheritDoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $libraries = array_keys($this->libraryManager->getLibrariesByExtension('web3_provider'));
    asort($libraries);

    $form['web3_js_library'] = [
      '#type' => 'item',
      '#title' => $this->t('Libraries'),
      '#description' => $this->t('Nothing to configure.') . '<br>' .
        $this->t("This plugin inject the following JavaScript libraries in order to interact with users' wallet:<br>@list", [
          '@list' => Markup::create('<code>' . implode('</code><code>', $libraries) . '</code>'),
        ]),
    ];

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function buildTransactionForm(array $form, FormStateInterface $form_state) {
    $form += parent::buildTransactionForm($form, $form_state);

    // $form['chainId']['#disabled'] = TRUE;
    // $form['from']['#disabled'] = TRUE;

    $form['chainId']['#wrapper_attributes']['id'] = 'wallet-network-wrapper';
    $form['chainId']['#attributes']['id'] = 'wallet-network';

    $form['from']['#wrapper_attributes']['id'] = 'wallet-from-wrapper';
    $form['from']['#attributes']['id'] = 'wallet-from';

    $form['to']['#attributes']['id'] = 'wallet-to';

    $form['value']['#attributes']['id'] = 'wallet-value';

    $ajax = ['callback' => [get_called_class(), 'ajaxSubmit']];

    $form['actions'] = ['#type' => 'actions'];

    $form['actions']['connect'] = [
      '#op' => 'eth_requestAccounts',
      '#type' => 'button',
      '#value' => $this->t('Connect'),
      '#ajax' => $ajax,
      '#weight' => 0,
      '#limit_validation_errors' => [],
      '#states' => [
        'visible' => [
          ':input#wallet-from' => ['filled' => FALSE]
        ],
      ],
    ];

    $form['actions']['disconnect'] = [
      '#op' => 'disconnect',
      '#type' => 'button',
      '#value' => $this->t('Reset'),
      '#ajax' => $ajax,
      '#weight' => 0,
      '#limit_validation_errors' => [],
      '#states' => [
        'invisible' => [
          ':input#wallet-from' => ['filled' => FALSE]
        ],
      ],
      '#attributes' => ['class' => ['button', 'button--danger']],
    ];

    $form['actions']['network'] = [
      '#op' => 'ethereum.networkVersion',
      '#type' => 'button',
      '#value' => $this->t('Select network'),
      '#ajax' => $ajax,
      '#weight' => 0,
      '#limit_validation_errors' => [],
      '#states' => [
        'disabled' => [
          ':input#wallet-from' => ['filled' => FALSE]
        ],
        'invisible' => [
          [':input#wallet-network' => ['filled' => TRUE]],
        ]
      ],
    ];

    $form['actions']['send'] = [
      '#op' => 'eth_sendTransaction',
      '#type' => 'button',
      '#value' => $this->t('Send'),
      '#ajax' => $ajax,
      '#weight' => 1,
      '#states' => [
        'disabled' => [
          [':input#wallet-from' => ['filled' => FALSE]],
          'or',
          [':input#wallet-to' => ['filled' => FALSE]],
          'or',
          [':input#wallet-value' => ['filled' => FALSE]],
        ],
      ],
    ];

    $form['#attached']['library'][] = 'core/drupal.ajax';
    $form['#attached']['library'][] = 'web3_provider/wallet.utils';
    $form['#attached']['library'][] = 'web3_provider/wallet.command';

    return $form;
  }

  /**
   * Ajax callback.
   */
  public static function ajaxSubmit(array $form, FormStateInterface $form_state) {
    if ($form_state instanceof SubformStateInterface) {
      $form_state->getCompleteFormState();
    }

    $response = new AjaxResponse();

    $action = $form_state->getTriggeringElement()['#op'] ?? NULL;

    if ($action == 'disconnect') {
      $form['web3_provider']['from']['#default_value'] = NULL;
      $form['web3_provider']['from']['#value'] = NULL;
      $response->addCommand(new ReplaceCommand('#wallet-from-wrapper', $form['web3_provider']['from']));

      $form_selector = '[data-drupal-selector="' . $form['#attributes']['data-drupal-selector'] . '"]';
      $response->addCommand(new ReplaceCommand($form_selector, $form));
      return $response;
    }

    // Catch results from Ajax command.
    $input = $form_state->getUserInput();
    if (($input['_triggering_element_name'] ?? NULL) == 'walletCommand') {
      $results = $input['_triggering_element_value'];
      if ($results['success'] ?? FALSE) {
        // Update sender field.
        if ($results['method'] == 'eth_requestAccounts') {
          $sender = $results['data'][0] ?? NULL;
          $form['web3_provider']['from']['#default_value'] = $sender;
          $form['web3_provider']['from']['#value'] = $sender;
          $response->addCommand(new ReplaceCommand('#wallet-from-wrapper', $form['web3_provider']['from']));
        }

        // Update chainId field.
        if ($results['method'] == 'ethereum.networkVersion') {
          $chaindId = $results['data'] ?? NULL;
          $form['web3_provider']['chainId']['#default_value'] = $chaindId;
          $form['web3_provider']['chainId']['#value'] = $chaindId;
          $response->addCommand(new ReplaceCommand('#wallet-network-wrapper', $form['web3_provider']['chainId']));
        }

        // Transaction submitted.
        if ($results['method'] == 'eth_sendTransaction') {
          $hash = $results['data'] ?? NULL;
          $form_selector = '[data-drupal-selector="' . $form['#attributes']['data-drupal-selector'] . '"]';
          $output = ['#markup' => $hash];
          $response->addCommand(new ReplaceCommand($form_selector, $output));
        }
      }
      return $response;
    }

    $values = $form_state->getValue('web3_provider');
    $transaction = new Transaction($values);
    $data = $transaction->data;
    $data['from'] = $values['from'];

    $response->addCommand(new WalletCommand($action, $data));
    $response->addCommand(new ScrollToCommand('.messages'));
    return $response;
  }
}
