<?php

namespace Drupal\web3_provider\Plugin\Web3Provider;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\PluginWithFormsTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\web3_provider\Web3ProviderPluginInterface;
use GuzzleHttp\RequestOptions;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for Web3Provider plugins.
 */
abstract class Web3ProviderBase extends PluginBase implements Web3ProviderPluginInterface, ContainerFactoryPluginInterface {

  use StringTranslationTrait;
  use PluginWithFormsTrait;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Key service.
   *
   * @var \Drupal\key\KeyRepositoryInterface
   */
  protected $keyRepository;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static($configuration, $plugin_id, $plugin_definition);
    $instance->logger = $container->get('logger.channel.web3_provider');
    $instance->httpClient = $container->get('http_client');
    $instance->eventDispatcher = $container->get('event_dispatcher');
    $instance->configFactory = $container->get('config.factory');
    $instance->keyRepository = $container->get('key.repository');

    // Populate with default values.
    $instance->setConfiguration($configuration);

    return $instance;
  }

  /*************************************************************************
   * Configurable plugin methods
   *************************************************************************/

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $plugin_id = $this->getPluginId();
    $config_key = 'web3_provider.provider.' . $plugin_id . '.settings';
    $plugin_config = $this->configFactory->get($config_key);

    return $plugin_config ? $plugin_config->getRawData() : [];
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $default_config = $this->defaultConfiguration();
    $this->configuration = $default_config + $configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /*************************************************************************
   * Configuration form.
   * @see \Drupal\web3_provider\PluginForm\Web3ProviderConfigFormBase
   *************************************************************************/

  /**
   * {@inheritDoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    // Nothing specific to validate by default.
  }

  /**
   * {@inheritDoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->setConfiguration($form_state->getValues());
  }

  /*************************************************************************
   * Transaction form.
   * @see \Drupal\web3_provider\PluginForm\Web3ProviderConfigFormBase
   *************************************************************************/

  /**
   * {@inheritDoc}
   */
  public function buildTransactionForm(array $form, FormStateInterface $form_state) {

    // ID of the targeted network as an integer.
    $networks = _web3_provider_get_ethereum_networks();

    $form['chainId'] = [
      '#type' => 'select',
      '#title' => $this->t('Network'),
      '#options' => $networks,
      '#required' => TRUE,
    ];

    $form['from'] = [
      // @todo Make this custom `address` element.
      '#type' => 'textfield',
      '#title' => $this->t('From'),
      // @todo Get sender from context.
      // '#default_value' => $context['wallet'],
      '#required' => TRUE,
    ];

    $form['to'] = [
      // @todo Make this custom `address` select element.
      '#type' => 'textfield',
      '#title' => $this->t('To'),
      // @todo Get contact list from context.
      // '#options' => $context['contacts'],
      '#required' => TRUE,
    ];

    $form['value'] = [
      // @todo Make this custom `value` element.
      '#type' => 'number',
      '#title' => $this->t('Value'),
      '#min' => 0,
      // @todo Get sender balance.
      // '#max' => $context['wallet']->getMax(),
      // '#validate' => ['::validateValue'],
      '#required' => TRUE,
      '#description' => $this->t('Integer of the value sent with this transaction.') . '<br>' .
        $this->t('Example: 10000000000 wei = 10 gwei = 0.00000001 ether.') . '<br>' .
        $this->t('Visit the @link for further information.', [
          '@link' => Link::fromTextAndUrl(
            $this->t('online converter'),
            Url::fromUri('https://eth-converter.com/', ['attributes' => [
              'target' => '_blank',
              'title' => $this->t('Open in a new tab'),
            ]]),
          )->toString()
        ]),
    ];

    $form['advanced'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Show advanced settings'),
      '#default_value' => 0,
      '#attributes' => ['id' => 'transaction-advanced-settings'],
    ];

    $show_advanced_state = [
      'visible' => [
        ':input#transaction-advanced-settings' => ['checked' => TRUE],
      ]
    ];

    // Nonce.
    $form['nonce'] = [
      '#type' => 'number',
      '#title' => $this->t('Nonce'),
      '#description' => $this->t('If empty, nonce value will be calculated automatically before sending the transaction'),
      '#min' => 1,
      '#states' => $show_advanced_state,
    ];

    // Miner tip (default: 1 wei).
    $form['maxPriorityFeePerGas'] = [
      '#type' => 'number',
      '#title' => $this->t('Max priority fee per gas'),
      '#description' => $this->t('e.g. Miner tip'),
      '#min' => 1,
      '#states' => $show_advanced_state,
    ];

    // maxFeePerGas = baseFeePerGas + maxPriorityFeePerGas
    $form['maxFeePerGas'] = [
      '#type' => 'number',
      '#title' => $this->t('Max fee per gas'),
      '#description' => $this->t('Speed up transaction.') . ' (maxFeePerGas = baseFeePerGas + maxPriorityFeePerGas)',
      '#states' => $show_advanced_state,
    ];

    // Estimated gas cost (default: 21000 wei which is the minimum for any ETH transaction).
    $form['gas'] = [
      '#type' => 'number',
      '#title' => $this->t('Estimated gas'),
      '#min' => 21000,
      '#states' => $show_advanced_state,
    ];

    // 'accessList' => $values['accessList'] ?? [],
    // 'data' => $values['data'] ?? '',

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function validateTransactionForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * {@inheritDoc}
   */
  public function submitTransactionForm(array &$form, FormStateInterface $form_state) {
  }

  /*************************************************************************
   * Web3ProviderPluginInterface methods.
   *************************************************************************/

  /**
   * {@inheritDoc}
   */
  public function getEndpointUri(): string {
    if ($key_id = $this->configuration['web3_http_provider'] ?? NULL) {
      $key = $this->keyRepository->getKey($key_id);
      $secret = $key->getKeyValues();
      return reset($secret) ?? NULL;
    }

    // Default Ganache CLI endpoint.
    return 'http://localhost:8545';
  }

  /**
   * {@inheritDoc}
   */
  public function requests(array $requests): array {
    $request_options = [];

    // @todo What headers would be useful?
    $request_options[RequestOptions::HEADERS] = [];

    // Process batch request.
    $i = 0;
    foreach (array_filter($requests) as $i => $request) {
      $evm_method = key($request);
      $parameters = $request[$evm_method];
      $batch[] = [
        'jsonrpc' => '2.0',
        'method' => $evm_method,
        'params' => $parameters,
        'id' => $i++,
      ];
    };

    $request_options[RequestOptions::BODY] = Json::encode($batch);

    $request_method = 'POST';
    $request_uri = $this->getEndpointUri();

    $this->logger->notice($this->t('Calling @method @uri:<br><pre>@body</pre>', [
      '@method' => $request_method,
      '@uri' => $request_uri,
      '@body' => $request_options[RequestOptions::BODY],
    ]));

    $response = $this->httpClient->request($request_method, $request_uri, $request_options);
    return Json::decode($response->getBody()->getContents());
  }
}
