<?php

/**
 * Helper class to prepare and sign EVM-compatible transactions.
 *
 * Inspired by web3p/ethereum-tx.
 *
 * @see https://github.com/web3p/ethereum-tx/blob/master/src/Transaction.php
 */

namespace Drupal\web3_provider\Utility;

use Web3p\EthereumTx\Transaction as Legacy;
use Web3p\EthereumTx\EIP2930Transaction;
use Web3p\EthereumTx\EIP1559Transaction;

/**
 * Prepare transaction to send to ETH networks.
 *
 * @see https://docs.infura.io/infura/networks/ethereum/concepts/transaction-types
 */
class Transaction {

  /**
   * @var Web3p\EthereumTx\TypeTransaction;
   */
  protected $tx;

  /**
   * The transaction type.
   *
   * @var int
   */
  public $type;

  /**
   * The provided values.
   *
   * @var array
   */
  public $values;

  /**
   * The transaction data.
   *
   * @var array
   */
  public $data;

  /**
   * Instanciate a new Transaction.
   *
   * @param array $values
   *   A given list of values.
   * @param int $type
   *   (optional) A given transaction type. Can be 0, 1 or 2 (default: 2).
   */
  public function __construct(array $values, int $type = 2) {
    $this->values = $values;
    $this->type = dechex($type);
    $this->data = $this->prepareData($values);

    // Need to decode chainId because of a bug in Web3php library.
    $params = $this->data;
    if (isset($params['chainId']) && (strpos($params['chainId'], '0x') === 0)) {
      $params['chainId'] = hexdec($params['chainId']);
    }

    switch ($type) {
      case '0x0':
        $this->tx = @new Legacy($params);
        break;
      case '0x1':
        $this->tx = @new EIP2930Transaction($params);
        break;
      default:
        $this->tx = @new EIP1559Transaction($params);
        break;
    }
  }

  /**
   * Help to clean values.
   *
   * @param array $values
   *   A given list of values.
   *
   * @return array
   *  A raw ethereum transaction data.
   */
  public function prepareData(array $values) {
    $minerTip = ($values['maxPriorityFeePerGas'] ?? NULL) ?: 1;
    $minerTip = strpos($minerTip, '0x') === 0 ? hexdec($minerTip) : $minerTip;

    $baseFeePerGas = ($values['baseFeePerGas'] ?? NULL) ?: 0;
    $baseFeePerGas = strpos($baseFeePerGas, '0x') === 0 ? hexdec($baseFeePerGas) : $baseFeePerGas;

    $maxFeePerGas = ($values['maxFeePerGas'] ?? NULL) ?: $baseFeePerGas;
    $maxFeePerGas += $minerTip;

    $data = [
      'nonce' => ($values['nonce'] ?? NULL) ?: 1,
      'from' => ($values['from'] ?? NULL) ?: '',
      'to' => ($values['to'] ?? NULL) ?: '',
      // Miner tip (default: 1 wei).
      'maxPriorityFeePerGas' => $minerTip ?? $maxFeePerGas,
      // Speed up transaction.
      'maxFeePerGas' => $maxFeePerGas,
      // Gas limit e.g estimated gas cost (default: 21000 wei which is the minimum for any ETH transaction).
      'gas' => ($values['gas'] ?? NULL) ?: 21000,
      'value' => ($values['value'] ?? NULL) ?: NULL,
      // Required. ID of the targeted network as an integer (default: 1337 Ganache e.g. local blockchain).
      'chainId' => ($values['chainId'] ?? NULL) ?: 1337,
      'accessList' => ($values['accessList'] ?? NULL) ?: [],
      'data' => ($values['data'] ?? NULL) ?: '',
    ];

    // Make sure all values as hex encoded.
    foreach ($data as $key => $value) {
      if (in_array($key, ['from', 'to', 'accessList', 'data'])) {
        continue;
      }
      $value = strpos($value, '0x') ? $value : (int) $value;
      $value = \is_int($value) ? dechex($value) : $value;
      $data[$key] = (strpos($value, '0x') !== 0) ? '0x' . $value : $value;
    }

    return $data;
  }

  /**
   * Update instance's values.
   *
   * @param array $values
   *   A given list of values.
   */
  public function refresh($values = []) {
    $this->data = $this->prepareData(array_merge($values, $this->data));
  }


  /**
   * Sign the transaction.
   *
   * @param string $private_key
   *  A given hex encoded private key.
   *
   * @return string
   *  The hex encoded signed ethereum transaction.
   */
  public function sign(string $private_key) {
    return '0x' . $this->tx->sign($private_key);
  }
}
